
from dataclasses import dataclass, asdict

from aio9p.helper import extract, mkfield

@dataclass
class Py9PStat:
    p9type: int # [2:4]
    p9dev: int # [4:8]
    p9qid: bytes # [8:21]
    p9mode: int # [21:25]
    p9atime: int # [25:29]
    p9mtime: int # [29:33]
    p9length: int # [33:41]
    p9name: bytes # [s]
    p9uid: bytes # [s]
    p9gid: bytes # [s]
    p9muid: bytes # [s]

    def size(self):
        return 49 + len(self.p9name) + len(self.p9uid) + len(self.p9gid) + len(self.p9muid)
    @staticmethod
    def from_stat(stat, qid):
        raise NotImplementedError
    @staticmethod
    def from_bytes(inpt, offset):
        nameoffset = offset + 41
        namelen = extract(inpt, nameoffset, 2)
        uidoffset = nameoffset + 2 + namelen
        uidlen = extract(inpt, uidoffset, 2)
        gidoffset = uidoffset + 2 + uidlen
        gidlen = extract(inpt, gidoffset, 2)
        muidoffset = gidoffset + 2 + gidlen
        muidlen = extract(inpt, muidoffset, 2)
        return Py9PStat(
            p9type=extract(inpt, offset+2, 2)
            , p9dev=extract(inpt, offset+4, 4)
            , p9qid=inpt[offset+8:offset+21]
            , p9mode=extract(inpt, offset+21, 4)
            , p9atime=extract(inpt, offset+25, 4)
            , p9mtime=extract(inpt, offset+29, 4)
            , p9length=extract(inpt, offset+33, 8)
            , p9name=inpt[nameoffset+2:nameoffset+2+namelen]
            , p9uid=inpt[uidoffset+2:uidoffset+2+uidlen]
            , p9gid=inpt[gidoffset+2:gidoffset+2+gidlen]
            , p9muid=inpt[muidoffset+2:muidoffset+2+muidlen]
            )
    def to_bytes(self, with_envelope=False):
        namelen = len(self.p9name)
        uidlen = len(self.p9uid)
        gidlen = len(self.p9gid)
        muidlen = len(self.p9muid)

        totallen = 49 + namelen + uidlen + gidlen + muidlen
        return b''.join((
            mkfield(totallen, 2) if with_envelope else b''
            , mkfield(totallen-2, 2) #Size field of the stat struct
            , mkfield(self.p9type, 2)
            , mkfield(self.p9dev, 4)
            , self.p9qid
            , mkfield(self.p9mode, 4)
            , mkfield(self.p9atime, 4)
            , mkfield(self.p9mtime, 4)
            , mkfield(self.p9length, 8)
            , mkfield(namelen, 2)
            , self.p9name
            , mkfield(uidlen, 2)
            , self.p9uid
            , mkfield(gidlen, 2)
            , self.p9gid
            , mkfield(muidlen, 2)
            , self.p9muid
            ))
    def to_dict(self):
        return asdict(self)




from asyncio import create_task, Protocol


import aio9p.constant as c
from aio9p.helper import extract, extract_bytefields, mkfield, mkbytefield, NULL_LOGGER
from aio9p.stat import Py9PStat

class Py9PException(Exception):
    pass

Py9PBadFID = Py9PException('Bad fid!')

class Py9PProtocol(Protocol):
    _logger = NULL_LOGGER
    def __init__(self, implementation, logger=None):
        if logger is not None:
            self._logger = logger
        self.maxsize = 256 #Default, overwritten by version negotiation
        self.implementation = implementation

        self._transport = None

        self._tasks = {}
        self._buffer = b''

        return None
    def connection_made(self, transport):
        self._logger.info('Connection made')
        self._transport = transport
        return None
    def connection_lost(self, exc):
        if exc is None:
            self._logger.info('Connection terminated')
        else:
            self._logger.info('Lost connection: %s', exc)
        return None
    def eof_received(self):
        self._logger.info('End of file received')
        return None
    def data_received(self, data):
        self._logger.debug('Data received: %s', data)
        buffer = self._buffer + data
        buflen = len(buffer)
        tasks = self._tasks
        msgstart = 0
        while msgstart < buflen - 7:
            msgsize = extract(buffer, msgstart, 4)
            msgend = msgstart + msgsize
            if buflen < msgend:
                break

            msgtype = extract(buffer, msgstart+4, 1)
            msgtag = buffer[msgstart+5:msgstart+7]

            if msgtype == c.TFLUSH:
                self.flush(msgtag, buffer[msgstart+7:msgstart+9])
                msgstart = msgend
                continue
            task = create_task(
                self.implementation.process_msg(msgtype, buffer[msgstart+7:msgend])
            )
            tasks[msgtag] = task
            task.add_done_callback(lambda x: self.sendmsg(msgtag, x))
            #Check if Flush
            #Create task - or: Chop out size, msgt, tag, and queue
            msgstart = msgend
        self._buffer = buffer[msgstart:]
        return None
    def flush(self, tag, oldtag):
        task = self._tasks.pop(oldtag, None)
        if task is None or task.cancelled():
            pass
        else:
            task.cancel()
        self._transport.writelines((
            mkfield(7, 7)
            , mkfield(c.RFLUSH, 1)
            , tag
            ))
        return None
    def sendmsg(self, msgtag, task):
        if task.cancelled():
            self._logger.debug('Sending message: cancelled task %s', msgtag)
            return None
        task_stored = self._tasks.pop(msgtag, None)
        if not task_stored == task:
            self._logger.debug('Sending message: Mismatched task %s', msgtag)
            raise ValueError(msgtag, task, task_stored)
        exception = task.exception()
        if exception is None:
            reslen, restype, fields = task.result()
        else:
            self._logger.info('Sending message: Got exception %s %s %s', msgtag, exception, task)
            reslen, restype, fields = self.implementation.errhandler(exception)
        res = (
            mkfield(reslen + 7, 4)
            , mkfield(restype, 1)
            , msgtag
            ) + fields
        binres = b''.join(res)
        self._logger.debug('Sending message: %s', binres.hex())
        self._transport.write(binres)
        self._logger.debug('Sent')
        return None


class Py9P2000():
    version = b'9P2000'
    _logger = NULL_LOGGER
    def __init__(self, maxsize, *args, logger=None, **kwargs):
        if logger is not None:
            self._logger = logger
        self.maxsize = maxsize
        return None
    def errhandler(self, exception):
        errstr = str(exception).encode(c.ENCODING)[:self.maxsize-9]
        self._logger.debug('Got exception: %s %s', errstr, exception)
        return c.RERROR, 2 + len(errstr), (mkfield(len(errstr), 2), errstr)
    async def process_msg(self, msgtype, msgbody):
        self._logger.debug('Processing: %s %s %s', msgtype, c.TRNAME.get(msgtype), msgbody.hex())
        dispatcher = {
            c.TVERSION: self._fmt_version
            , c.TATTACH: self._fmt_attach
            , c.TSTAT: self._fmt_stat
            , c.TCLUNK: self._fmt_clunk
            , c.TWALK: self._fmt_walk
            , c.TOPEN: self._fmt_open
            , c.TREAD: self._fmt_read
            , c.TWRITE: self._fmt_write
            , c.TCREATE: self._fmt_create
            , c.TWSTAT: self._fmt_wstat
            , c.TREMOVE: self._fmt_remove
            }
        processor = dispatcher.get(msgtype)
        if processor is None:
            self._logger.debug('Unknown message type %s', msgtype)
            raise NotImplementedError
        res = await processor(msgbody)
        self._logger.debug('Replying with message: %s', res)
        return res


    async def _fmt_version(self, msgbody):
        #TODO: Abort outstanding IO
        maxsize = extract(msgbody, 0, 4)
        versionlength = extract(msgbody, 4, 2)
        if len(msgbody) < 6 + versionlength:
            self._logger.error('Message body too short for version string')
            raise ValueError(msgbody)
        version = msgbody[6:6+versionlength]
        if version != self.version: #TODO: Check behaviour here
            srvverlen, srvverfields = mkbytefield(b'unknown')
            return 4 + srvverlen, c.RVERSION, (mkfield(256, 4), *srvverfields)
        srvmax = await self.maxsize_backend(maxsize)
        self.maxsize = min(maxsize, srvmax)
        srvverlen, srvverfields = mkbytefield(self.version)
        return 4 + srvverlen, c.RVERSION, (
            srvmax.to_bytes(4, 'little')
            , *srvverfields
            )
    async def maxsize_backend(self, client_maxsize):
        '''
        Calculates the backend's maximum message size. The implementation will
        transmit the minimum of client and backend size.

        The default here calculates the smallest power of two less than the
        client maximum. It can and probably should be overridden.
        '''
        res = 1024
        while res < client_maxsize:
            res = res << 1
        return res >> 1

    async def _fmt_attach(self, msgbody):
        fid = msgbody[0:4]
        afid = msgbody[4:8]
        unamelen = extract(msgbody, 8, 2)
        uname = msgbody[10:10+unamelen]

        anamelen = extract(msgbody, 10+unamelen, 2)
        aname = msgbody[10+unamelen:12+anamelen]

        qid = await self.attach(fid, afid, uname, aname)

        return 13, c.RATTACH, (qid,)
    async def attach(self, fid, afid, uname, aname):
        raise NotImplementedError

    async def _fmt_auth(self, msgbody):
        afid = msgbody[0:4]
        unamelen = extract(msgbody, 4, 2)
        uname = msgbody[6:6+unamelen]

        anamelen = extract(msgbody, 6+unamelen, 2)
        aname = msgbody[8+unamelen:8+unamelen+anamelen]

        aqid = await self.auth(afid, uname, aname)

        return 13, c.RAUTH, (aqid,)
    async def auth(self, afid, uname, aname):
        raise NotImplementedError

    async def _fmt_stat(self, msgbody):
        fid = msgbody[0:4]
        statres = await self.stat(fid)
        statbytes = statres.to_bytes(with_envelope=True)
        return len(statbytes), c.RSTAT, (statbytes,)
    async def stat(self, fid):
        raise NotImplementedError

    async def _fmt_clunk(self, msgbody):
        fid = msgbody[0:4]
        await self.clunk(fid)
        return 0, c.RCLUNK, ()
    async def clunk(self, fid):
        raise NotImplementedError

    async def _fmt_walk(self, msgbody):
        fid = msgbody[0:4]
        newfid = msgbody[4:8]
        count = extract(msgbody, 8, 2)
        try:
            wnames = extract_bytefields(msgbody, 10, count)
        except ValueError:
            self._logger.debug('Could not build walknames: Count %s, %s', count, msgbody[10:].hex())
            raise
        qids = await self.walk(fid, newfid, wnames)
        if count and not qids:
            errmsg = b'No such file!'
            errenvelope = mkfield(13, 2)
            return 15, c.RERROR, (errenvelope, errmsg)
        qidcount = len(qids)
        return 2 + 13*qidcount, c.RWALK, (mkfield(qidcount, 2),) + qids
    async def walk(self, fid, newfid, wnames):
        raise NotImplementedError

    async def _fmt_open(self, msgbody):
        fid = msgbody[0:4]
        mode = extract(msgbody, 4, 1)
        qid, iounit = await self.open(fid, mode)
        return 17, c.ROPEN, (qid, mkfield(iounit, 4))
    async def open(self, fid, mode):
        raise NotImplementedError

    async def _fmt_read(self, msgbody):
        fid = msgbody[0:4]
        offset = extract(msgbody, 4, 8)
        count = extract(msgbody, 12, 4)
        resdata = await self.read(fid, offset, count)
        #TODO: Length overflow checking
        resdatalen = len(resdata)
        return 4 + resdatalen, c.RREAD, (mkfield(resdatalen, 4), resdata)
    async def read(self, fid, offset, count):
        raise NotImplementedError

    async def _fmt_write(self, msgbody):
        fid = msgbody[0:4]
        offset = extract(msgbody, 4, 8)
        count = extract(msgbody, 12, 4)
        #TODO: Check that count matches the size
        data = msgbody[16:16+count]
        rescount = await self.write(fid, offset, data)
        return 4, c.RWRITE, (mkfield(rescount, 4),)
    async def write(self, fid, offset, data):
        raise NotImplementedError

    async def _fmt_create(self, msgbody):
        fid = msgbody[0:4]
        namelen = extract(msgbody, 4, 2)
        name = msgbody[6:6+namelen]

        perm = extract(msgbody, 6+namelen, 4)
        mode = extract(msgbody, 10+namelen, 1)
        qid, iounit = await self.create(fid, name, perm, mode)
        return 17, c.RCREATE, (qid, mkfield(iounit, 4))
    async def create(self, fid, name, perm, mode):
        raise NotImplementedError

    async def _fmt_wstat(self, msgbody):
        fid = msgbody[0:4]
        stat = Py9PStat.from_bytes(msgbody, 6)
        await self.wstat(fid, stat)
        return 0, c.RWSTAT, ()
    async def wstat(self, fid, stat):
        raise NotImplementedError

    async def _fmt_remove(self, msgbody):
        fid = msgbody[0:4]
        await self.remove(fid)
        return 0, c.RREMOVE, ()
    async def remove(self, fid):
        raise NotImplementedError




class Implementation9P():
    def errhandler(self, exception):
        # Returns reslen, restype, (field, field, ...)
        raise NotImplementedError

    async def version(self, maxsize, version):
        # Returns maxsize, version
        raise NotImplementedError
    async def attach(self, fid, afid, uname, aname, *args):
        # 9P2000.u : args = (n_uname,)
        # Returns root qid
        raise NotImplementedError
    async def auth(self, afid, uname, aname, *args):
        # 9P2000.u : args = (n_uname,)
        raise NotImplementedError
    async def walk(self, fid, newfid, wnames):
        raise NotImplementedError
    async def read(self, fid, offset, count):
        raise NotImplementedError
    async def write(self, fid, offset, count, data):
        raise NotImplementedError
    async def clunk(self, fid):
        raise NotImplementedError
    async def remove(self, fid):
        raise NotImplementedError


    #9P2000.L
    async def statfs(self, fid):
        raise NotImplementedError
    async def lopen(self, fid, openflags):
        raise NotImplementedError
    async def lcreate(self, fid, name, createflags, mode, gid):
        raise NotImplementedError
    async def symlink(self, fid, name, symtgt, gid):
        raise NotImplementedError
    async def mknod(self, dfid, name, mode, major, minor, gid):
        raise NotImplementedError
    async def readlink(self, fid):
        raise NotImplementedError
    async def getattr(self, fid, reqmask):
        raise NotImplementedError
    async def setattr(self, fid, valid, mode, uid, gid, size, atime_sec, atime_nsec, mtime_sec, mtime_nsec):
        raise NotImplementedError
    async def xattrwalk(self, fid, newfid, name):
        raise NotImplementedError
    async def xattrcreate(self, fid, attr_size, flags):
        #Attention: The actual setxattr should happen when fid is clunked
        raise NotImplementedError
    async def readdir(self, fid, offset, count):
        raise NotImplementedError
    async def fsync(self, fid):
        raise NotImplementedError
    async def lock(self, fid, locktype, flags, start, length, proc_id, client_id):
        raise NotImplementedError
    async def getlock(self, fid, locktype, start, length, proc_id, client_id):
        raise NotImplementedError
    async def link(self, dfid, fid, name):
        raise NotImplementedError
    async def mkdir(self, dfid, name, mode, gid):
        raise NotImplementedError
    async def renameat(self, olddirfid, oldname, newdirfd, newname):
        raise NotImplementedError
    async def unlinkat(self, dirfid, name, flags):
        raise NotImplementedError












from itertools import chain
from logging import getLogger, NullHandler

from aio9p.constant import ENCODING

NULL_LOGGER = getLogger('')
NULL_LOGGER.setLevel('DEBUG')
NULL_LOGGER.addHandler(NullHandler())

def mkqid(mode, base, version=0):
    if isinstance(base, int):
        base = base.to_bytes(8, 'little')
    return b''.join((
        (mode >> 24).to_bytes(1, 'little')
        , version.to_bytes(4, 'little')
        , base
        ))

def extract(msg, offset, size):
    return int.from_bytes(msg[offset:offset+size], byteorder='little')

def extract_bytefields(msg, offset, count):
    return tuple(_gen_bytefields(msg, offset, count))

def _gen_bytefields(msg, offset, count):
    msglen = len(msg)
    while count > 0:
        if msglen < offset + 2:
            raise ValueError('Incomplete length field', msg, offset)
        fieldlen = extract(msg, offset, 2)
        nextoffset = offset + 2 + fieldlen
        if msglen < nextoffset:
            raise ValueError('Incomplete content field', msg, offset)
        yield msg[offset+2:nextoffset]
        offset = nextoffset
        count = count - 1


def mkfield(value, size):
    try:
        return value.to_bytes(size, byteorder='little')
    except OverflowError as e:
        raise ValueError from e

def mkbytefield(*payloads): #TODO: Rename to mkbytefields
    total = sum(2 + len(payload) for payload in payloads)
    resfields = tuple(chain.from_iterable(
        (len(payload).to_bytes(2, byteorder='little'), payload)
        for payload in payloads
        ))
    return total, resfields

def mkstrfield(*args): #TODO: Rename to mkstrfields
    return mkbytefield(*(
        value.encode(ENCODING)
        for value in args
        ))




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aio9p', 'aio9p.example']

package_data = \
{'': ['*']}

install_requires = \
['pytest>=7.1.3,<8.0.0']

setup_kwargs = {
    'name': 'aio9p',
    'version': '0.1.0',
    'description': '',
    'long_description': '# aio9p\n\nAsyncio-based bindings for the 9P protocol. Work in progress, littered with print-debugs. A working example for the 9P2000 dialect is implemented in aio9p.example.simple .\n\n# TODO\n\n- Support for 9P2000.u and 9P2000.L dialects\n- Separation between protocol parser-formatters and the classes used for implementation subclassing\n- Logging\n- Significantly more tests\n',
    'author': 'florp',
    'author_email': 'jens.krewald@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)




from asyncio import run, get_running_loop
from logging import getLogger, StreamHandler, Formatter

from aio9p.protocol import Py9PProtocol

async def example_server(implementation):
    print('Example server running')
    loop = get_running_loop()
    logger = getLogger('example-9p')
    handler = StreamHandler()
    fmt = Formatter('Logger: %(message)s')
    logger.setLevel('DEBUG')
    handler.setLevel('DEBUG')
    handler.setFormatter(fmt)
    logger.addHandler(handler)
    server = await loop.create_server(
        lambda: Py9PProtocol(implementation(1024, logger=logger), logger=logger)
        , '127.0.0.1'
        , 8090
        )
    async with server:
        await server.serve_forever()

def example_main(implementation):
    run(example_server(implementation))



ENCODING = 'utf-8'

# QID types
QTDIR = 1 << 7
QTAPPEND = 1 << 6
QTEXCL = 1 << 5
QTHISTORIC = 1 << 4
QTAUTH = 1 << 3
QTTMP = 1 << 2
QTSYMLINK = 1 << 1
QTLINK = 1 << 0
QTFILE = 0

# QID types shifted for use with access modes
DMDIR = QTDIR << 24
DMAPPEND = QTAPPEND << 24
DMEXCL = QTEXCL << 24
DMHISTORIC = QTHISTORIC << 24
DMAUTH = QTAUTH << 24
DMTMP = QTTMP << 24
DMSYMLINK = QTSYMLINK << 24
DMLINK = QTLINK << 24
DMFILE = QTFILE << 24

# QID types as single bytes
QTByteDIR = QTDIR.to_bytes(1, 'little')
QTByteAPPEND = QTAPPEND.to_bytes(1, 'little')
QTByteEXCL = QTEXCL.to_bytes(1, 'little')
QTByteAUTH = QTAUTH.to_bytes(1, 'little')
QTByteTMP = QTTMP.to_bytes(1, 'little')
QTByteSYMLINK = QTSYMLINK.to_bytes(1, 'little')
QTByteLINK = QTLINK.to_bytes(1, 'little')
QTByteFILE = QTFILE.to_bytes(1, 'little')

# Message types

TLERROR = 6
RLERROR = (TLERROR + 1)
TSTATFS = 8
RSTATFS = (TSTATFS + 1)
TLOPEN = 12
RLOPEN = (TLOPEN + 1)
TLCREATE = 14
RLCREATE = (TLCREATE + 1)
TSYMLINK = 16
RSYMLINK = (TSYMLINK + 1)
TMKNOD = 18
RMKNOD = (TMKNOD + 1)
TRENAME = 20
RRENAME = (TRENAME + 1)
TREADLINK = 22
RREADLINK = (TREADLINK + 1)
TGETATTR = 24
RGETATTR = (TGETATTR + 1)
TSETATTR = 26
RSETATTR = (TSETATTR + 1)
TXATTRWALK = 30
RXATTRWALK = (TXATTRWALK + 1)
TXATTRCREATE = 32
RXATTRCREATE = (TXATTRCREATE + 1)
TREADDIR = 40
RREADDIR = (TREADDIR + 1)
TFSYNC = 50
RFSYNC = (TFSYNC + 1)
TLOCK = 52
RLOCK = (TLOCK + 1)
TGETLOCK = 54
RGETLOCK = (TGETLOCK + 1)
TLINK = 70
RLINK = (TLINK + 1)
TMKDIR = 72
RMKDIR = (TMKDIR + 1)
TRENAMEAT = 74
RRENAMEAT = (TRENAMEAT + 1)
TUNLINKAT = 76
RUNLINKAT = (TUNLINKAT + 1)
TVERSION = 100
RVERSION = (TVERSION + 1)
TAUTH = 102
RAUTH = (TAUTH + 1)
TATTACH = 104
RATTACH = (TATTACH + 1)
TERROR = 106
RERROR = (TERROR + 1)
TFLUSH = 108
RFLUSH = (TFLUSH + 1)
TWALK = 110
RWALK = (TWALK + 1)
TOPEN = 112
ROPEN = (TOPEN + 1)
TCREATE = 114
RCREATE = (TCREATE + 1)
TREAD = 116
RREAD = (TREAD + 1)
TWRITE = 118
RWRITE = (TWRITE + 1)
TCLUNK = 120
RCLUNK = (TCLUNK + 1)
TREMOVE = 122
RREMOVE = (TREMOVE + 1)
TSTAT = 124
RSTAT = (TSTAT + 1)
TWSTAT = 126
RWSTAT = (TWSTAT + 1)


TRNAME = {
    TLERROR: "TLERROR"
    , RLERROR: "RLERROR"
    , TSTATFS: "TSTATFS"
    , RSTATFS: "RSTATFS"
    , TLOPEN: "TLOPEN"
    , RLOPEN: "RLOPEN"
    , TLCREATE: "TLCREATE"
    , RLCREATE: "RLCREATE"
    , TSYMLINK: "TSYMLINK"
    , RSYMLINK: "RSYMLINK"
    , TMKNOD: "TMKNOD"
    , RMKNOD: "RMKNOD"
    , TRENAME: "TRENAME"
    , RRENAME: "RRENAME"
    , TREADLINK: "TREADLINK"
    , RREADLINK: "RREADLINK"
    , TGETATTR: "TGETATTR"
    , RGETATTR: "RGETATTR"
    , TSETATTR: "TSETATTR"
    , RSETATTR: "RSETATTR"
    , TXATTRWALK: "TXATTRWALK"
    , RXATTRWALK: "RXATTRWALK"
    , TXATTRCREATE: "TXATTRCREATE"
    , RXATTRCREATE: "RXATTRCREATE"
    , TREADDIR: "TREADDIR"
    , RREADDIR: "RREADDIR"
    , TFSYNC: "TFSYNC"
    , RFSYNC: "RFSYNC"
    , TLOCK: "TLOCK"
    , RLOCK: "RLOCK"
    , TGETLOCK: "TGETLOCK"
    , RGETLOCK: "RGETLOCK"
    , TLINK: "TLINK"
    , RLINK: "RLINK"
    , TMKDIR: "TMKDIR"
    , RMKDIR: "RMKDIR"
    , TRENAMEAT: "TRENAMEAT"
    , RRENAMEAT: "RRENAMEAT"
    , TUNLINKAT: "TUNLINKAT"
    , RUNLINKAT: "RUNLINKAT"
    , TVERSION: "TVERSION"
    , RVERSION: "RVERSION"
    , TAUTH: "TAUTH"
    , RAUTH: "RAUTH"
    , TATTACH: "TATTACH"
    , RATTACH: "RATTACH"
    , TERROR: "TERROR"
    , RERROR: "RERROR"
    , TFLUSH: "TFLUSH"
    , RFLUSH: "RFLUSH"
    , TWALK: "TWALK"
    , RWALK: "RWALK"
    , TOPEN: "TOPEN"
    , ROPEN: "ROPEN"
    , TCREATE: "TCREATE"
    , RCREATE: "RCREATE"
    , TREAD: "TREAD"
    , RREAD: "RREAD"
    , TWRITE: "TWRITE"
    , RWRITE: "RWRITE"
    , TCLUNK: "TCLUNK"
    , RCLUNK: "RCLUNK"
    , TREMOVE: "TREMOVE"
    , RREMOVE: "RREMOVE"
    , TSTAT: "TSTAT"
    , RSTAT: "RSTAT"
    , TWSTAT: "TWSTAT"
    , RWSTAT: "RWSTAT"
    }




from errno import ENOENT
from os import strerror

from aio9p.constant import QTByteDIR, DMDIR, QTByteFILE, DMFILE, RERROR, ENCODING
from aio9p.helper import mkbytefield, mkstrfield, mkqid, mkfield
from aio9p.protocol import Py9PException, Py9P2000, Py9PException, Py9PBadFID
from aio9p.stat import Py9PStat

from aio9p.example import example_main

BASEQID = QTByteDIR + b'\x00\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00'
FILEQID = QTByteFILE + b'\x00\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00'

BASENAME = b'/'
FILENAME = b'barfile'
FILECONTENT = b'Hello, barfilereader'

BASESTAT = Py9PStat(
    p9type=0
    , p9dev=0
    , p9qid=BASEQID
    , p9mode=(DMDIR | 0o777)
    , p9atime=0
    , p9mtime=0
    , p9length=0
    , p9name=BASENAME
    , p9uid=b'root'
    , p9gid=b'root'
    , p9muid=b'root'
    )

def filestat(fc):
    return Py9PStat(
        p9type=0
        , p9dev=0
        , p9qid=FILEQID
        , p9mode=(DMFILE | 0o777)
        , p9atime=0
        , p9mtime=0
        , p9length=len(fc)
        , p9name=FILENAME
        , p9uid=b'root'
        , p9gid=b'root'
        , p9muid=b'root'
        )

def mkfilestat(qid, qname, qlen):
    return Py9PStat(
        p9type=0
        , p9dev=0
        , p9qid=qid
        , p9mode=(DMFILE | 0o777)
        , p9atime=0
        , p9mtime=0
        , p9length=qlen
        , p9name=qname
        , p9uid=b'root'
        , p9gid=b'root'
        , p9muid=b'root'
        )

def mkdirstat(qid, qname):
    return Py9PStat(
        p9type=0
        , p9dev=0
        , p9qid=qid
        , p9mode=(DMDIR | 0o777)
        , p9atime=0
        , p9mtime=0
        , p9length=0
        , p9name=qname
        , p9uid=b'root'
        , p9gid=b'root'
        , p9muid=b'root'
        )

def mkstatlen(name):
    return 49 + len(name) + 4 + 4 + 4

class Simple9P(Py9P2000):
    def __init__(self, maxsize, logger=None):
        super().__init__(maxsize, logger=logger)
        self._logger.info('Simple9P running! Version:', self.version)
        self._fid = {}
        self._name = {
            BASEQID: BASENAME
            , FILEQID: FILENAME
            }
        self._dircontent = {
            BASEQID: {
                FILENAME: FILEQID
                }
            }
        self._content = {
            FILEQID: FILECONTENT
            }
        return None
    def errhandler(self, exception):
        self._logger.debug('Exception: %s', exception)
        if isinstance(exception, Py9PException):
            msg = exception.args[0]
            self._logger.debug('Py9PException:', msg)
        else:
            msg = f'Exception: {exception}'
        if isinstance(msg, int):
            errstr = strerror(msg)
            self._logger.debug(f'Integer exception %i, message %s', msg, errstr)
            errstrlen, errstrfields = mkstrfield(errstr)
            errnofield = mkfield(msg, 4)
            return errstrlen + 4, RERROR, (*errstrfields, errnofield)
        bytemsg = str(msg).encode(ENCODING)
        msgfieldslen, msgfields = mkbytefield(bytemsg)
        return msgfieldslen, RERROR, msgfields
    async def maxsize_backend(self, client_maxsize):
        return 65535
    async def attach(self, fid, afid, uname, aname):
        self._fid[fid] = BASEQID
        return BASEQID
    async def auth(self, afid, uname, aname):
        self._logger.error('Attempted auth: %s %s %s', afid, uname, aname)
        raise NotImplementedError
    async def stat(self, fid):
        qid = self._fid.get(fid)
        name = self._name.get(qid)
        if name is None:
            raise Py9PBadFID
        content = self._content.get(qid)
        if content is not None:
            return mkfilestat(qid, name, len(content))
        return mkdirstat(qid, name)
    async def clunk(self, fid):
        self._fid.pop(fid, None)
        return None
        
    async def walk(self, fid, newfid, wnames):
        self._logger.debug('Simple walk from %s to %s: %s', fid, newfid, wnames)
        if not wnames:
            oldqid = self._fid.get(fid)
            if oldqid is None:
                raise Py9PBadFID
            self._fid[newfid] = oldqid
            return ()
        dirqid = self._fid.get(fid)
        wqids = []
        for wname in wnames:
            dircontent = self._dircontent.get(dirqid, {})
            wqid = dircontent.get(wname)
            if wqid is None:
                break
            wqids.append(wqid)
            dirqid = wqid
        if not wqids:
            raise Py9PException(ENOENT)
        if len(wqids) == len(wnames):
            self._fid[newfid] = wqids[-1]
        return tuple(wqids)
    async def open(self, fid, mode):
        # if mode in {1, 2}:
        #     raise Py9PException('Files are read-only')
        qid = self._fid.get(fid)
        if qid is None:
            raise Py9PBadFID
        return qid, 0
    async def read(self, fid, offset, count):
        qid = self._fid.get(fid)
        if qid is None:
            self._logger.error('Bad Read FID: %s %s', fid, self._fid)
            raise Py9PBadFID
        content = self._content.get(qid)
        if content is not None:
            return content[offset:offset+count]
        dircontent = self._dircontent.get(qid)
        if dircontent is None:
            raise Py9PBadFID
        diroffset = 0
        for entryname, entryqid in sorted(dircontent.items()):
            if diroffset == offset:
                entrycontent = self._content.get(entryqid)
                if entrycontent is None:
                    return mkdirstat(entryqid, entryname).to_bytes()
                return mkfilestat(entryqid, entryname, len(entrycontent)).to_bytes()
            if diroffset > offset:
                raise Py9PException('Bad directory read offset')
            diroffset = diroffset + mkstatlen(entryname)
        return b''
    async def write(self, fid, offset, data):
        #TODO: Check that fid is okay
        qid = self._fid.get(fid)
        fc = self._content.get(qid)
        if fc is None: #TODO: Writes to directories should give different error
            self._logger.error('Bad Write FID: %s %s', fid, self._fid)
            raise Py9PBadFID
        self._logger.debug('Writing to %s at %i for length %i : %s ', qid, offset, len(data), data)
        if len(fc) < offset:
            return 0
        self._content[qid] = fc[:offset] + data + fc[offset+len(data):]
        return len(data)
    async def create(self, fid, name, perm, mode):
        qid = self._fid.get(fid)
        dircontent = self._dircontent.get(qid)
        if dircontent is None:
            self._logger.error('Bad Create FID: %s %s', fid, self._fid)
            raise Py9PBadFID
        if name in dircontent.keys():
            self._logger.error('Bad Create filename: %s %s %s', name, qid, dircontent)
            raise Py9PException(f'File name exists: {name}')
        newqid = mkqid(
            mode
            , int.from_bytes(max(k[5:] for k in self._name.keys()), 'little') + 1
            )
        dircontent[name] = newqid
        self._name[newqid] = name
        if perm & DMDIR:
            self._dircontent[newqid] = {}
        else:
            self._content[newqid] = b''
        self._fid[fid] = newqid
        return newqid, 0
    async def wstat(self, fid, stat):
        return None
    async def remove(self, fid):
        qid = self._fid.pop(fid, None)
        if qid is None:
            return None
        filename = self._name.pop(qid, None)
        for dirc in self._dircontent.values():
            if dirc.get(filename) == qid:
                dirc.pop(filename)
        self._dircontent.pop(qid, None)
        self._content.pop(qid, None)
        return None


if __name__ == "__main__":
    example_main(Simple9P)

